`timescale 1ns / 1ps

//******************************************************************************
//* MiniRISC CPU v2.0                                                          *
//*                                                                            *
//* Aritmetikai-logikai egysg (ALU). Tmogatott mveletek:                    *
//* 1) Adatmozgats (nincs mveletvgzs)                                      *
//* 2) sszeads s kivons tvitel nlkl, illetve tvitellel                 *
//* 3) Bitenknti S, VAGY s XOR                                              *
//* 4) Shiftels s forgats                                                   *
//******************************************************************************
module alu(
   //rajel.
   input  wire       clk,
   
   //Vezrl bemenetek.
   input  wire [1:0] op_type,       //Az ALU mvelet tpust kivlaszt jel
   input  wire [1:0] arith_sel,     //Az aritmetikai mveletet kivlaszt jel
   input  wire [1:0] logic_sel,     //A logikai mveletet kivlaszt jel
   input  wire [3:0] shift_sel,     //A shiftelsi mveletet kivlaszt jel
   
   //Az operandusok s az eredmny.
   input  wire [7:0] operand1,      //Els operandus
   input  wire [7:0] operand2,      //Msodik operandus
   output wire [7:0] result,        //Az ALU mvelet eredmnye
   
   //ALU felttel jelek.
   input  wire [3:0] flag_din,      //A flag-ekbe rand rtk
   input  wire       flag_wr,       //A flag-ek rs engedlyez jele
   output reg        flag_z,        //Zero flag
   output reg        flag_c,        //Carry flag
   output reg        flag_n,        //Negative flag
   output reg        flag_v         //Overflow flag
);

`include "src\MiniRISC_CPU\control_defs.vh"
`include "src\MiniRISC_CPU\opcode_defs.vh"

//******************************************************************************
//* Az aritmetikai mveletek vgrehajtsa (sszeads s kivons).              *
//*                                                                            *
//* Kivons esetn a 2. operandus kettes komplemenst (/op2 + 1) adjuk hozz   *
//* az 1. operandushoz. sszead/kivon logikt nem tudunk itt hasznlni, mert *
//* szksges az tvitel bemenet s kimenet is.                                *
//******************************************************************************

//Az arith_sel[1] bit hatrozza meg az elvgzend mveletet:
//0: sszeads
//1: kivons (kettes komplemens hozzadsa)
wire [7:0] adder_op2 = operand2 ^ {8{arith_sel[1]}};

//Az arith_sel[0] bit vlasztja ki a C flag hasznlatt:
//0: mveletvgzs tvitel nlkl
//1: mveletvgzs tvitellel (kivons esetn negljuk)
wire carry_in  = (arith_sel[0]) ? (flag_c ^ arith_sel[1]) : arith_sel[1];

//Az sszeads elvgzse. Az eredmny MSb-je a kimeneti tvitel bit.
wire [8:0] sum = operand1 + adder_op2 + carry_in;

//Kivons esetn a kimen tvitel bitet negljuk.
wire [8:0] arith_result = {sum[8] ^ arith_sel[1], sum[7:0]};

//A kettes komplemens tlcsorduls szlelse. Tlcsorduls akkor trtnik,
//ha azonos eljel szmokat adunk ssze s az sszeg eljele ettl eltr.
wire overflow = (~operand1[7] & ~adder_op2[7] &  arith_result[7]) |
                ( operand1[7] &  adder_op2[7] & ~arith_result[7]);


//******************************************************************************
//* A logikai mveletek vgrehajtsa (bitenknti S, VAGY, XOR). Itt hajtjuk   *
//* vgre az als s a fels 4 bit felcserlst is.                           *
//******************************************************************************
reg [7:0] logic_result;

always @(*)
begin
   case (logic_sel)
      //Bitenknti S
      LOGIC_AND: logic_result <= operand1 & operand2;
      //Bitenknti VAGY
      LOGIC_OR : logic_result <= operand1 | operand2;
      //Bitenknti XOR
      LOGIC_XOR: logic_result <= operand1 ^ operand2;
      //Az als s a fels 4 bit felcserlse
      default  : logic_result <= {operand1[3:0], operand1[7:4]};
   endcase
end


//******************************************************************************
//* A shiftelsi s a forgatsi mveletek elvgzse.                           *
//******************************************************************************
reg [8:0] shift_result;

//Forgats esetn a shift_sel[2] vlasztja ki a belptetend bitet:
//0: sima forgats, a kilp bit kerl belptetsre a msik oldalon
//1: forgats a carry flag-en keresztl
wire rot_in_l = (shift_sel[2]) ? flag_c : operand1[7];
wire rot_in_r = (shift_sel[2]) ? flag_c : operand1[0];

//A shift_sel[3] vlasztja ki a shiftels tpust:
//0: norml shiftels
//1: aritmetikai shiftels, az MSb (eljel bit) helyben marad
wire shr_msb  = (shift_sel[3]) ? operand1[7] : shift_sel[2];

//A shiftels/forgats elvgzse. Az eredmny MSb-je a kimeneti tvitel bit.
always @(*)
begin
   case (shift_sel[1:0])
      //Shiftels balra
      SHIFT_SHL: shift_result <= {operand1[7], operand1[6:0], shift_sel[2]};
      //Shiftels jobbra
      SHIFT_SHR: shift_result <= {operand1[0], shr_msb, operand1[7:1]};
      //Forgats balra
      SHIFT_ROL: shift_result <= {operand1[7], operand1[6:0], rot_in_l};
      //Forgats jobbra
      SHIFT_ROR: shift_result <= {operand1[0], rot_in_r, operand1[7:1]};
   endcase
end


//******************************************************************************
//* Az ALU eredmnyt kivlaszt multiplexer.                                   *
//******************************************************************************
reg [8:0] alu_result;

always @(*)
begin
   case (op_type)
      ALU_MOVE : alu_result <= {1'b0, operand2};         //Adatmozgats
      ALU_ARITH: alu_result <= arith_result;             //Aritmetikai mvelet
      ALU_LOGIC: alu_result <= {1'b0, logic_result};     //Logikai mvelet
      ALU_SHIFT: alu_result <= shift_result;             //Shiftels/forgats
   endcase
end

//A multiplexer kimenetnek als 8 bitje az ALU eredmny, az MSb
//a carry flag j rtkt tartalmazza.
assign result = alu_result[7:0];


//******************************************************************************
//* Zero flag (Z).                                                             *
//*                                                                            *
//* Jelzi, ha az ALU mvelet eredmnye 0. rtkt frissteni kell aritmetikai, *
//* logikai vagy shiftelsi/forgatsi mveletek vgrehajtsa esetn. rhatnak *
//* kell lennie, mert megszaktsbl val visszatrsnl vissza kell lltani  *
//* a korbbi rtkt.                                                         *
//******************************************************************************
always @(posedge clk)
begin
   if (flag_wr)
      flag_z <= flag_din[Z_FLAG];
   else
      if (op_type != ALU_MOVE)
         flag_z <= (result == 8'd0);
end


//******************************************************************************
//* Carry flag (C).                                                            *
//*                                                                            *
//* Jelzi, ha az aritmetikai mveletek sorn tvitel trtnt, illetve felveszi *
//* a kishiftelt bit rtkt shiftels/forgats vgrehajtsa esetn. rhatnak *
//* kell lennie, mert megszaktsbl val visszatrsnl vissza kell lltani  *
//* a korbbi rtkt.                                                         *
//******************************************************************************
always @(posedge clk)
begin
   if (flag_wr)
      flag_c <= flag_din[C_FLAG];
   else
      if ((op_type == ALU_ARITH) || (op_type == ALU_SHIFT))
         flag_c <= alu_result[8];
end


//******************************************************************************
//* Negative flag (N).                                                         *
//*                                                                            *
//* Jelzi, ha az ALU mvelet eredmnye negatv (az eljel bit 1). rhatnak    *
//* kell lennie, mert megszaktsbl val visszatrsnl vissza kell lltani  *
//* a korbbi rtkt.                                                         *
//******************************************************************************
always @(posedge clk)
begin
   if (flag_wr)
      flag_n <= flag_din[N_FLAG];
   else
      if (op_type != ALU_MOVE)
         flag_n <= alu_result[7];
end


//******************************************************************************
//* Overflow flag (V).                                                         *
//*                                                                            *
//* Jelzi, ha kettes komplemens tlcsorduls trtnt az aritmetikai mveletek  *
//* vgrehajtsa esetn. rhatnak kell lennie, mert megszaktsbl val       *
//* visszatrsnl vissza kell lltani a korbbi rtkt.                     *
//******************************************************************************
always @(posedge clk)
begin
   if (flag_wr)
      flag_v <= flag_din[V_FLAG];
   else
      if (op_type == ALU_ARITH)
         flag_v <= overflow;
end


endmodule
