`timescale 1ns / 1ps

//******************************************************************************
//* MiniRISC CPU v2.0                                                          *
//*                                                                            *
//* A processzort vezrl llapotgp.                                          *
//******************************************************************************
module controller_fsm(
   //rajel s reset.
   input  wire       clk,              //rajel
   input  wire       rst,              //Aszinkron reset
   
   //A lehvott utastssal kapcsolatos jelek.
   input  wire       addr_op2_sel,     //Az ALU 2. operandusnak kivlasztsa
   input  wire [3:0] opcode,           //Az utastsban lv mveleti kd
   input  wire [3:0] ctrl_op,          //A programvezrlsi mvelet kdja
   
   //A processzor llapotval kapcsolatos jelek.
   output wire       initialize,       //Inicializls
   output reg        fetch,            //Utasts lehvs
   output wire       interrupt,        //Megszakts kiszolgls
   
   //A programvezrl utastsok vgrehajtsval kapcsolatos jelek.
   output reg        ex_jump,          //Ugrs vgrehajtsa
   output wire       ex_call,          //Szubrutinhvs vgrehajtsa
   output wire       ex_ret_sub,       //Visszatrs szubrutinbl
   output wire       ex_ret_int,       //Visszatrs megszaktsbl
   
   //Az adatstruktrval kapcsolatos jelek.
   output wire       wr_data_sel,      //A regiszterbe rand adat kivlasztsa
   output reg        reg_wr_en,        //A regisztertmb rs engedlyez jele
   output reg  [1:0] alu_op_type,      //ALU mvelet kivlaszt jel
   input  wire       alu_flag_z,       //Zero flag
   input  wire       alu_flag_c,       //Carry flag
   input  wire       alu_flag_n,       //Negative flag
   input  wire       alu_flag_v,       //Overflow flag
   
   //Az adatmemrival kapcsolatos jelek.
   output wire       data_mem_wr,      //rs engedlyez jel
   output wire       data_mem_rd,      //Olvass engedlyez jel
   
   //A megszaktssal kapcsolatos jelek.
   input  wire       irq,              //Megszaktskr bemenet
   input  wire       flag_ie_din,      //Az IE flag-ba rand rtk
   output reg        flag_ie,          //Megyszakts engedlyez flag (IE)
   input  wire       flag_if_din,      //Az IF flag-ba rand rtk
   output reg        flag_if,          //Megyszakts flag (IF)
   
   //A debug interfsz jelei.
   input  wire       dbg_break,        //Program vgrehajtsnak felfggesztse
   input  wire       dbg_continue,     //Program vgrehajtsnak folytatsa
   input  wire       dbg_ie_wr,        //Az IE flag rs engedlyez jele
   input  wire       dbg_ie_din,       //Az IE bitbe rand adat
   input  wire       dbg_reg_wr,       //A regisztertmb rs engedlyez jele
   input  wire       dbg_mem_wr,       //Az adatmemria rs engedlyez jele
   input  wire       dbg_mem_rd,       //Az adatmemria olvass engedlyez jele
   output wire       dbg_is_brk        //A trspont llapot jelzse
);

`include "src\MiniRISC_CPU\control_defs.vh"
`include "src\MiniRISC_CPU\opcode_defs.vh"

//******************************************************************************
//* Megyszakts engedlyez flag (IE).                                        *
//******************************************************************************
wire ie_clr;
wire ie_set;

always @(posedge clk)
begin
   if (initialize || interrupt || ie_clr)
      flag_ie <= 1'b0;
   else
      if (ie_set)
         flag_ie <= 1'b1;
      else
         if (ex_ret_int)
            flag_ie <= flag_ie_din;
         else
            if (dbg_is_brk && dbg_ie_wr)
               flag_ie <= dbg_ie_din;
end


//******************************************************************************
//* Megyszakts flag (IF). A megszakts kiszolglst jelzi.                 *
//******************************************************************************
always @(posedge clk)
begin
   if (initialize)
      flag_if <= 1'b0;
   else
      if (interrupt)
         flag_if <= 1'b1;
      else
         if (ex_ret_int)
            flag_if <= flag_if_din;
end


//******************************************************************************
//* A vezrl llapotgp.                                                      *
//******************************************************************************
localparam STATE_INIT     = 4'd0;   //Inicializls
localparam STATE_FETCH    = 4'd1;   //Utasts lehvs
localparam STATE_DECODE   = 4'd2;   //Utasts dekdols
localparam STATE_EX_LD    = 4'd3;   //Utasts vgrehajts (memria olvass)
localparam STATE_EX_ST    = 4'd4;   //Utasts vgrehajts (memria rs)
localparam STATE_EX_MOV   = 4'd5;   //Utasts vgrehajts (adatmozgats)
localparam STATE_EX_ARITH = 4'd6;   //Utasts vgrehajts (aritmetikai)
localparam STATE_EX_LOGIC = 4'd7;   //Utasts vgrehajts (logikai/csere)
localparam STATE_EX_SHIFT = 4'd8;   //Utasts vgrehajts (shiftels/forgats)
localparam STATE_EX_CTRL  = 4'd9;   //Utasts vgrehajts (programvezrls)
localparam STATE_EX_NOP   = 4'd10;  //Utasts vgrehajts (nincs mveletvgzs)
localparam STATE_INT_REQ  = 4'd11;  //Megszaktskrs kezelse
localparam STATE_BREAK    = 4'd12;  //Trspont

//Az aktulis llapotot trol regiszter.
reg [3:0] state;

always @(posedge clk or posedge rst)
begin
   if (rst)
      state <= STATE_INIT;
   else
      case (state)
         //Inicializls.
         STATE_INIT    : state <= STATE_FETCH;
         
         //Utasts lehvs.
         STATE_FETCH   : if (dbg_break)
                            state <= STATE_BREAK;
                         else
                            state <= STATE_DECODE;
         
         //Utasts dekdols.
         STATE_DECODE  : case (opcode)
                            //Adatmemria olvass.
                            OPCODE_LD   : state <= STATE_EX_LD;
                            
                            //Adatmemria rs.
                            OPCODE_ST   : state <= STATE_EX_ST;
                            
                            //Adatmozgats vagy konstans betlts.
                            OPCODE_MOV  : state <= STATE_EX_MOV;
                            
                            //Aritmetikai mveletek.
                            OPCODE_ADD  : state <= STATE_EX_ARITH;
                            OPCODE_ADC  : state <= STATE_EX_ARITH;
                            OPCODE_SUB  : state <= STATE_EX_ARITH;
                            OPCODE_SBC  : state <= STATE_EX_ARITH;
                            OPCODE_CMP  : state <= STATE_EX_ARITH;
                            
                            //Logikai mveletek.
                            OPCODE_AND  : state <= STATE_EX_LOGIC;
                            OPCODE_OR   : state <= STATE_EX_LOGIC;
                            OPCODE_XOR  : state <= STATE_EX_LOGIC;
                            OPCODE_TST  : state <= STATE_EX_LOGIC;
                            
                            //Shiftels/forgats vagy csere.
                            OPCODE_SHIFT: if (addr_op2_sel)
                                             state <= STATE_EX_SHIFT;
                                          else
                                             state <= STATE_EX_LOGIC;
                            
                            //Programvezrls.
                            OPCODE_CTRL : state <= STATE_EX_CTRL;
                            
                            //Nincs mveletvgzs.
                            default     : state <= STATE_EX_NOP;
                         endcase
         
         //Utasts vgrehajts.
         STATE_EX_LD   : if (flag_ie && irq)
                            state <= STATE_INT_REQ;
                         else
                            state <= STATE_FETCH;
                            
         STATE_EX_ST   : if (flag_ie && irq)
                            state <= STATE_INT_REQ;
                         else
                            state <= STATE_FETCH;
         
         STATE_EX_MOV  : if (flag_ie && irq)
                            state <= STATE_INT_REQ;
                         else
                            state <= STATE_FETCH;
         
         STATE_EX_ARITH: if (flag_ie && irq)
                            state <= STATE_INT_REQ;
                         else
                            state <= STATE_FETCH;
         
         STATE_EX_LOGIC: if (flag_ie && irq)
                            state <= STATE_INT_REQ;
                         else
                            state <= STATE_FETCH;
         
         STATE_EX_SHIFT: if (flag_ie && irq)
                            state <= STATE_INT_REQ;
                         else
                            state <= STATE_FETCH;
         
         STATE_EX_CTRL : if (flag_ie && irq)
                            state <= STATE_INT_REQ;
                         else
                            state <= STATE_FETCH;
         
         STATE_EX_NOP  : if (flag_ie && irq)
                            state <= STATE_INT_REQ;
                         else
                            state <= STATE_FETCH;
                            
         //Megszaktskrs kezelse.
         STATE_INT_REQ : state <= STATE_FETCH;
         
         //Trspont.
         STATE_BREAK   : if (dbg_continue)
                            state <= STATE_DECODE;
                         else
                            state <= STATE_BREAK;
         
         //rvnytelen llapotok.
         default       : state <= STATE_INIT;
      endcase
end


//******************************************************************************
//* A processzor llapotval kapcsolatos jelek.                                *
//******************************************************************************
//Inicializls.
assign initialize = (state == STATE_INIT);

//Utasts lehvs.
always @(*)
begin
   case (state)
      STATE_FETCH: fetch <= ~dbg_break;
      STATE_BREAK: fetch <= dbg_continue;
      default    : fetch <= 1'b0;
   endcase
end

//Megszakts kiszolglsa.
assign interrupt  = (state == STATE_INT_REQ);

//Trspont.
assign dbg_is_brk = (state == STATE_BREAK);


//******************************************************************************
//* A programvezrl utastsokkal kapcsolatos jelek.                          *
//******************************************************************************
//Felttel nlkli s feltteles ugrs jelzse.
always @(*)
begin
   if (state == STATE_EX_CTRL)
      case (ctrl_op)
         CTRL_JMP: ex_jump <= 1'b1;
         CTRL_JZ : ex_jump <= alu_flag_z;
         CTRL_JNZ: ex_jump <= ~alu_flag_z;
         CTRL_JC : ex_jump <= alu_flag_c;
         CTRL_JNC: ex_jump <= ~alu_flag_c;
         CTRL_JN : ex_jump <= alu_flag_n;
         CTRL_JNN: ex_jump <= ~alu_flag_n;
         CTRL_JV : ex_jump <= alu_flag_v;
         CTRL_JNV: ex_jump <= ~alu_flag_v;
         default : ex_jump <= 1'b0;
      endcase
   else
      ex_jump <= 1'b0;
end

//Szubrutinhvs jelzse.
assign ex_call    = (state == STATE_EX_CTRL) & (ctrl_op == CTRL_JSR);

//Visszatrs szubrutinbl.
assign ex_ret_sub = (state == STATE_EX_CTRL) & (ctrl_op == CTRL_RTS);

//Visszatrs megszaktsbl.
assign ex_ret_int = (state == STATE_EX_CTRL) & (ctrl_op == CTRL_RTI);

//A megszakts engedlyez flag vezrl jelei.
assign ie_set     = (state == STATE_EX_CTRL) & (ctrl_op == CTRL_STI);
assign ie_clr     = (state == STATE_EX_CTRL) & (ctrl_op == CTRL_CLI);


//******************************************************************************
//* Az adatstruktra vezrl jelei.                                            *
//******************************************************************************
//A regiszertmbbe rand adat kivlaszt jele:
//0: ALU mvelet eredmnye
//1: az adatmemribl olvasott adat
assign wr_data_sel = (state == STATE_EX_LD);

//Regisztertmb rs engedlyez jel.
always @(*)
begin
   case (state)
      STATE_EX_LD   : reg_wr_en <= 1'b1;
      STATE_EX_MOV  : reg_wr_en <= 1'b1;
      STATE_EX_ARITH: reg_wr_en <= ~opcode[3];
      STATE_EX_LOGIC: reg_wr_en <= ~opcode[3];
      STATE_EX_SHIFT: reg_wr_en <= 1'b1;
      STATE_BREAK   : reg_wr_en <= dbg_reg_wr;
      default       : reg_wr_en <= 1'b0;
   endcase
end
   
//ALU mvelet kivlasz jel.
always @(*)
begin
   case (state)
      STATE_EX_ARITH: alu_op_type <= ALU_ARITH;
      STATE_EX_LOGIC: alu_op_type <= ALU_LOGIC;
      STATE_EX_SHIFT: alu_op_type <= ALU_SHIFT;
      default       : alu_op_type <= ALU_MOVE;
   endcase
end


//******************************************************************************
//* Az adatmemrival kapcsolatos jelek.                                       *
//******************************************************************************
//Az adatmemria rs engedlyez jele.
assign data_mem_wr = (dbg_is_brk) ? dbg_mem_wr : (state == STATE_EX_ST);

//Az adatmemria olvass engedlyez jele.
assign data_mem_rd = (dbg_is_brk) ? dbg_mem_rd : (state == STATE_EX_LD);

   
endmodule
