`timescale 1ns / 1ps

//******************************************************************************
//* Egyszer 8 bites idzt perifria.                                        *
//*                                                                            *
//* Az itzt peridusideje T = (TR + 1) * PS * Tclk, ahol TR az idzt      *
//* szmll kezdllapota (0 - 255), PS az eloszts (1, 16, 64, 256, 1024,   *
//* 4096, 16384 vagy 65536) s Tclk a rendszerrajel peridusideje. A mkdsi *
//* md lehet egyszeres vagy peridikus, ismtelt jratltssel. Az idztsi  *
//* peridus vgn lehetsg van megszaktskrsre is.                        *
//*                                                                            *
//* A perifria cme paramter tadssal llthat be a fels szint modulban  *
//* az idzt megpldnyostsakor. A szintzis mr ennek megfeleleen az     *
//* aktulis BASEADDR bziscmmel trtnik.                                    *

//* A programozi fellet:                                                     *
//*                                                                            *
//* Cm         Tpus   Bitek                                                  *
//* BASEADDR+0  WR      Az idzt szmll kezdllapota                      *
//*                     TR7   TR6   TR5   TR4   TR3   TR2   TR1   TR0          *
//* BASEADDR+0  RD      Az idzt szmll aktulis rtke                    *
//*                     TM7   TM6   TM5   TM4   TM3   TM2   TM1   TM0          *
//* BASEADDR+1  WR      Parancs regiszter                                      *
//*                     TIE   TPS2  TPS1  TPS0  -     -     TREP  TEN          *
//* BASEADDR+1  RD      Sttusz regiszter                                      *
//*                     TIT   TPS2  TPS1  TPS0  0     TOUT  TREP  TEN          *
//*                                                                            *
//* A parancs regiszter bitjei statikus set/clear rtelemben hasznlhatk. A   *
//* sttusz regiszter md bitjei statikusak, az esemnyjelz TOUT (s ezzel    * 
//* egytt az esetleg aktv IRQ is) olvassra trldik.                        *
//******************************************************************************
module basic_timer #(
   //A perifria bziscme.
   parameter BASEADDR = 8'hff
) (
   //rajel s reset.
   input  wire       clk,              //rajel
   input  wire       rst,              //Reset jel
   
   //A busz interfsz jelei.
   input  wire [7:0] cpu2dmem_addr,    //Cmbusz
   input  wire       cpu2dmem_wr,      //rs engedlyez jel
   input  wire       cpu2dmem_rd,      //Olvass engedlyez jel
   input  wire [7:0] cpu2dmem_data,    //rsi adatbusz
   output reg  [7:0] dmem2cpu_data,    //Olvassi adatbusz
   
   //Megszaktskr kimenet.
   output wire       irq
);

//******************************************************************************
//* Cmdekdols.                                                              *
//******************************************************************************
//A perifria kivlaszt jele.
wire psel = ((cpu2dmem_addr >> 1) == (BASEADDR >> 1));

//A szmll kezdllapot regiszter rsnak jelzse.
wire tr_wr = psel & cpu2dmem_wr & ~cpu2dmem_addr[0];

//A parancs regiszter rsnak jelzse.
wire tc_wr = psel & cpu2dmem_wr &  cpu2dmem_addr[0];

//A szmll olvassnak jelzse.
wire tm_rd = psel & cpu2dmem_rd & ~cpu2dmem_addr[0];

//A sttusz regiszter olvassnak jelzse.
wire ts_rd = psel & cpu2dmem_rd &  cpu2dmem_addr[0];


//******************************************************************************
//* Az idzt szmlljnak kezdllapott trol regiszter.                  *
//******************************************************************************
reg [7:0] tr_reg;

always @(posedge clk)
begin
   if (rst)
      tr_reg <= 8'hff;                       //Reset: maximum peridus belltsa
   else
      if (tr_wr)
         tr_reg <= cpu2dmem_data;            //A kezdllapot regiszter rsa   
end


//******************************************************************************
//* A parancs regiszter bitjeit egyedileg specifikljuk.                       *
//******************************************************************************
//Az eloszt s a szmll vgllapot jelzsei.
wire       ps_tc, tmr_cnt_tc;
//A parancs regiszter bitjei.
reg        ten, trep, tie;
reg  [2:0] tps;

always @(posedge clk)
begin
   if (rst)
      ten <= 1'b0;                           //Reset: az idzt tiltsa
   else
      if (tc_wr)
         ten <= cpu2dmem_data[0];            //Parancs regiszter rs
      else
         if (ps_tc && tmr_cnt_tc && !trep)   //Ha egyszeri zemmdban lejrt,
            ten <= 1'b0;                     //akkor az engedlyezst trljk
end

always @(posedge clk)
begin
   if (rst)
      trep <= 1'b0;                          //Reset: egyszeri zemmd belltsa
   else
      if (tc_wr)
         trep <= cpu2dmem_data[1];           //Parancs regiszter rs
end

always @(posedge clk)
begin
   if (rst)
      tps <= 3'd0;                           //Reset: 1 eloszts
   else
      if (tc_wr)
         tps <= cpu2dmem_data[6:4];          //Parancs regiszter rs
end

always @(posedge clk)
begin
   if (rst)
      tie <= 1'b0;                           //Reset: megszakts tiltsa
   else
      if (tc_wr)
         tie <= cpu2dmem_data[7];            //Parancs regiszter rs
end


//******************************************************************************
//* Az eloszt.                                                               *
//******************************************************************************
//Az eloszts rtknek kivlasztsa.
reg [15:0] ps_val;

always @(*)
begin
   case (tps)
      3'b000: ps_val <= 16'd0;               //     1 eloszts
      3'b001: ps_val <= 16'd15;              //    16 eloszts
      3'b010: ps_val <= 16'd63;              //    64 eloszts
      3'b011: ps_val <= 16'd255;             //   256 eloszts
      3'b100: ps_val <= 16'd1023;            //  1024 eloszts
      3'b101: ps_val <= 16'd4095;            //  4096 eloszts
      3'b110: ps_val <= 16'd16383;           // 16384 eloszts
      3'b111: ps_val <= 16'd65535;           // 65536 eloszts
   endcase
end

//Az eloszt szmllja.
reg [15:0] ps_cnt;
reg        ps_cnt_clr;

assign ps_tc = (ps_cnt == 16'd0);            //Az eloszt vgllapotnak jelzse

always @(posedge clk)
begin
   ps_cnt_clr <= tc_wr;
end

always @(posedge clk)
begin
   if (rst)
      ps_cnt <= 16'd0;                       //Reset: 1 eloszts
   else
      if (ps_cnt_clr)
         ps_cnt <= ps_val;                   //Parancs regiszter rs: jratlts
      else
         if (ten)
            if (ps_tc)
               ps_cnt <= ps_val;             //Vgllapot: jratlts
            else
               ps_cnt <= ps_cnt - 16'd1;     //Egybknt lefele szmll
end
      

//******************************************************************************
//* Az idzt egy 8 bites lefel szmll.                                    *
//******************************************************************************     
reg [7:0] tmr_cnt;      

assign tmr_cnt_tc = (tmr_cnt == 8'h00);      //A nulla vgrtket jelezzk

always @ (posedge clk) 
begin                     
   if (rst)
      tmr_cnt <= 8'hff;                      //Reset: max. rtk betltse
   else 
      if (tr_wr)
         tmr_cnt <= cpu2dmem_data;           //j kezdrtk betltse kzvetlenl
      else
         if (ten && ps_tc)                   //Ha az idzt engedlyezett s az
            if (tmr_cnt_tc)                  //eloszt vgllapotban van:
               tmr_cnt <= tr_reg;            // -idzts vgn, 0-nl jratlts   
            else
               tmr_cnt <= tmr_cnt - 8'd1;    // -egybknt szmol lefel
end


//******************************************************************************
//* A sttusz regiszter bitjei.                                                *
//****************************************************************************** 
reg tout;
     
//Idzts lejrt jelzs a sttusz regiszterben.
//A bellts nagyobb priorits, mint a trls.
always @ (posedge clk) 
begin                     
   if (rst)
      tout <= 1'b0;                          //Reset: a TOUT jelzs trlse
   else 
      if (ps_tc && tmr_cnt_tc && ten)        //Ha az idzt engedlyezett s
         tout <= 1'b1;                       //elrte a vgrtket, akkor jelzs 
      else
         if (ts_rd)
            tout <= 1'b0;                    //A sttusz reg. olvassa trli
end

//Megszaktskrs az idzts lejrtakor, ha engedlyezett.
assign irq = tie & tout;

//A 8 bites sttusz informci sszelltsa az egyedi bitek alapjn.
//Az IRQ s a TOUT bit majdnem azonos jelents, de az IRQ bit magba
//foglalja az ITEN llapott is. 
wire [7:0] status_reg = {irq, tps, 1'b0, tout, trep, ten};


//******************************************************************************
//* Az olvassi adatbusz meghajtsa.                                           *
//****************************************************************************** 
always @(*)
begin
   case ({ts_rd, tm_rd})
      2'b01  : dmem2cpu_data <= tmr_cnt;     //Idzt szmllja
      2'b10  : dmem2cpu_data <= status_reg;  //Sttusz regiszter
      default: dmem2cpu_data <= 8'd0;        //Inaktv 0 rtk, ha nincs olvass
   endcase
end

endmodule
