`timescale 1ns / 1ps

//******************************************************************************
//* Soros kommunikcit biztost slave USRT perifria.                        *
//*                                                                            *
//* Az USRT kommunikci keretezett formtumot hasznl: 1 START bit (0),       *
//* 8 adatbit (LSb elszr) s 1 STOP bit (1), a soros adatvonal inaktv       *
//* szintje 1. Az USRT rajel hatroza meg az adattviteli sebessget, az      *
//* rajelet a master egysg adja ki a slave egysg fel. Az ad az USRT       *
//* rajel felfut lre adja ki a kvetkez bitet, melyet a vev a lefut     *
//* lre mintavtelez. A vev csak a kerethiba mentes karaktereket trolja     *
//* (STOP bit = 1), a hibs karaktereket eldobja.                              *
//*                                                                            *
//* A perifria cme paramter tadssal llthat be a fels szint modulban  *
//* az USRT modul megpldnyostsakor. A szintzis mr ennek megfeleleen az  *
//* aktulis BASEADDR bziscmmel trtnik.                                    *
//*                                                                            *
//* A programozi fellet:                                                     *
//*                                                                            *
//* Cm         Tpus   Bitek                                                  *
//* BASEADDR+0  WR      Vezrl regiszter                                      *
//*                     -     -     RXCLR TXCLR RXIE  TXIE  RXEN  TXEN         *
//* BASEADDR+0  RD      Sttusz regiszter                                      *
//*                     RXNE  TXNF  0     0     RXIE  TXIE  RXEN  TXEN         *
//* BASEADDR+1  WR      Az adsi FIFO rsa (ha TXNF=1)                        *
//* BASEADDR+1  RD      A vteli FIFO olvassa (ha RXNE=1)                     *
//*                     D7    D6    D5    D4    D3    D2    D1    D0           *
//******************************************************************************
module slave_usrt #(
   //A perifria bziscme.
   parameter BASEADDR = 8'hff
) (
   //rajel s reset.
   input  wire       clk,              //rajel
   input  wire       rst,              //Reset jel
   
   //A soros interfsz jelei.
   input  wire       usrt_clk,         //USRT rajel
   input  wire       usrt_rxd,         //Soros adatbemenet
   output wire       usrt_txd,         //Soros adatkimenet
   
   //A busz interfsz jelei.
   input  wire [7:0] cpu2dmem_addr,    //Cmbusz
   input  wire       cpu2dmem_wr,      //rs engedlyez jel
   input  wire       cpu2dmem_rd,      //Olvass engedlyez jel
   input  wire [7:0] cpu2dmem_data,    //rsi adatbusz
   output reg  [7:0] dmem2cpu_data,    //Olvassi adatbusz
   
   //Megszaktskr kimenet.
   output wire       irq
);

//******************************************************************************
//* Cmdekdols.                                                              *
//******************************************************************************
//A perifria kivlaszt jele.
wire psel = ((cpu2dmem_addr >> 1) == (BASEADDR >> 1));

//A kontroll regiszter rsnak jelzse.
wire ctrl_reg_wr = psel & ~cpu2dmem_addr[0] & cpu2dmem_wr;

//Az adatregiszter rsnak jelzse.
wire data_reg_wr = psel &  cpu2dmem_addr[0] & cpu2dmem_wr;

//A sttusz regiszter olvassnak jelzse.
wire stat_reg_rd = psel & ~cpu2dmem_addr[0] & cpu2dmem_rd;

//Az adatregiszter olvassnak jelzse.
wire data_reg_rd = psel &  cpu2dmem_addr[0] & cpu2dmem_rd;


//******************************************************************************
//* A kontroll regiszter.                                                      *
//******************************************************************************
reg [3:0] ctrl_reg;

always @(posedge clk)
begin
   if (rst)
      ctrl_reg <= 4'b0000;
   else
      if (ctrl_reg_wr)
         ctrl_reg <= cpu2dmem_data[3:0];
end

//A kontroll regiszter bitjei.
wire tx_enable   = ctrl_reg[0];
wire rx_enable   = ctrl_reg[1];
wire tx_int_en   = ctrl_reg[2];
wire rx_int_en   = ctrl_reg[3];
wire tx_fifo_clr = rst | (ctrl_reg_wr & cpu2dmem_data[4]);
wire rx_fifo_clr = rst | (ctrl_reg_wr & cpu2dmem_data[5]);


//******************************************************************************
//* A sttusz regiszter.                                                       *
//******************************************************************************
wire [7:0] stat_reg;
wire       rx_fifo_empty;
wire       tx_fifo_full;

//A sttusz regiszter bitjei.
assign stat_reg[3:0] = ctrl_reg;
assign stat_reg[5:4] = 2'd0;
assign stat_reg[6]   = ~tx_fifo_full;
assign stat_reg[7]   = ~rx_fifo_empty;


//******************************************************************************
//* Az olvassi adatbusz s a megszaktskr kimenet meghajtsa.              *
//******************************************************************************
wire [7:0] rx_data;

always @(*)
begin
   case ({data_reg_rd, stat_reg_rd})
      2'b01  : dmem2cpu_data <= stat_reg;
      2'b10  : dmem2cpu_data <= rx_data;
      default: dmem2cpu_data <= 8'd0;
   endcase
end

assign irq = (tx_int_en & ~tx_fifo_full) | (rx_int_en & ~rx_fifo_empty);


//******************************************************************************
//* ldetektls az USRT rajelen s az RXD vonal ksleltetse.                *
//******************************************************************************
(* shreg_extract = "no" *)
(* register_balancing = "no" *)
(* register_duplication = "no" *)
(* equivalent_register_removal = "no" *)
reg  [2:0] usrt_clk_samples;
wire       usrt_clk_rising  = (usrt_clk_samples[2:1] == 2'b01);
wire       usrt_clk_falling = (usrt_clk_samples[2:1] == 2'b10);

always @(posedge clk)
begin
   usrt_clk_samples <= {usrt_clk_samples[1:0], usrt_clk};
end

(* shreg_extract = "no" *)
(* register_balancing = "no" *)
(* register_duplication = "no" *)
(* equivalent_register_removal = "no" *)
reg  [1:0] rxd_samples;

always @(posedge clk)
begin
   rxd_samples <= {rxd_samples[0], usrt_rxd};
end


//******************************************************************************
//* USRT vteli llapotgp.                                                    *
//******************************************************************************
localparam RX_WAIT_START = 2'd0;
localparam RX_DATA_BITS  = 2'd1;
localparam RX_STOP_BIT   = 2'd2;
localparam RX_FIFO_WRITE = 2'd3;

//Az aktulis llapotot trol regiszter.
reg [1:0] rx_state;
//Shiftregiszter a vett bitek trolshoz.
reg [7:0] rx_shr;
//A vett bitek szmt trol regiszter.
reg [2:0] rx_bit_cnt;

always @(posedge clk)
begin
   if (rst)
      rx_state <= RX_WAIT_START;
   else
      case (rx_state)                  
         //Vrakozs a START bitre.
         RX_WAIT_START: begin
                           rx_bit_cnt <= 3'd0;
                            
                           if (rx_enable && usrt_clk_falling && (rxd_samples[1] == 0))
                              rx_state <= RX_DATA_BITS;
                           else
                              rx_state <= RX_WAIT_START;
                        end
                         
         //A 8 adatbit vtele.
         RX_DATA_BITS : if (usrt_clk_falling)
                        begin
                           rx_shr     <= {rxd_samples[1], rx_shr[7:1]};
                           rx_bit_cnt <= rx_bit_cnt + 3'd1;
                            
                           if (rx_bit_cnt == 3'd7)
                              rx_state <= RX_STOP_BIT;
                           else
                              rx_state <= RX_DATA_BITS;
                        end
                        else
                           rx_state <= RX_DATA_BITS;
                         
         //A STOP bit vtele.
         RX_STOP_BIT  : if (usrt_clk_falling)
                           if (rxd_samples[1])
                              rx_state <= RX_FIFO_WRITE;
                           else
                              rx_state <= RX_WAIT_START;
                        else
                           rx_state <= RX_STOP_BIT;
                            
         //A vett adat bersa a FIFO-ba.
         RX_FIFO_WRITE: rx_state <= RX_WAIT_START;
      endcase
end


//******************************************************************************
//* A vteli FIFO.                                                             *
//******************************************************************************
wire rx_fifo_wr = (rx_state == RX_FIFO_WRITE);

fifo rx_fifo(
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rx_fifo_clr),                  //Reset jel
   
   //Adatvonalak.
   .data_in(rx_shr),                   //A FIFO-ba rand adat
   .data_out(rx_data),                 //A FIFO-bl kiolvasott adat
   
   //Vezrl bemenetek.
   .write(rx_fifo_wr),                 //rs engedlyez jel
   .read(data_reg_rd),                 //Olvass engedlyez jel
   
   //Sttusz kimenetek.
   .empty(rx_fifo_empty),              //A FIFO res
   .full()                             //A FIFO tele van
);


//******************************************************************************
//* USRT adsi llapotgp.                                                     *
//******************************************************************************
localparam TX_WAIT_ENABLE = 2'd0;
localparam TX_CHECK_FIFO  = 2'd1;
localparam TX_READ_FIFO   = 2'd2;
localparam TX_SEND_DATA   = 2'd3;

//Az aktulis llapotot trol regiszter.
reg  [1:0] tx_state;
//Az adsi shiftregiszter.
reg  [8:0] tx_shr;
//Az elkldtt bitek szmt trol regiszter.
reg  [3:0] tx_bit_cnt;
//Az adsi FIFO-bl beolvasott adat.
wire [7:0] tx_fifo_dout;
//Az adsi FIFO res jelzse.
wire       tx_fifo_empty;

always @(posedge clk)
begin
   if (rst)
   begin
      tx_shr   <= 9'b1_1111_1111;
      tx_state <= TX_WAIT_ENABLE;
   end
   else
      case (tx_state)
         //Vrakozs, amg az ad tiltva van.
         TX_WAIT_ENABLE: begin
                            tx_shr <= 9'b1_1111_1111;
                            
                            if (tx_enable)
                               tx_state <= TX_CHECK_FIFO;
                            else
                               tx_state <= TX_WAIT_ENABLE;
                         end
                            
         //Vrakozs, amg az adsi FIFO res.
         TX_CHECK_FIFO : if (usrt_clk_rising && (tx_fifo_empty == 0))
                            tx_state <= TX_READ_FIFO;
                         else
                            tx_state <= TX_CHECK_FIFO;
                         
         //Adat olvassa a FIFO-bl s bersa a shiftregiszterbe.
         TX_READ_FIFO  : begin
                            tx_shr     <= {tx_fifo_dout, 1'b0};
                            tx_bit_cnt <= 4'd0;
                            
                            tx_state   <= TX_SEND_DATA;
                         end
                         
         //Az adat elkldse.
         TX_SEND_DATA  : if (usrt_clk_rising)
                         begin
                            tx_shr     <= {1'b1, tx_shr[8:1]};
                            tx_bit_cnt <= tx_bit_cnt + 4'd1;
                            
                            if (tx_bit_cnt[3])
                               tx_state <= TX_WAIT_ENABLE;
                            else
                               tx_state <= TX_SEND_DATA;
                         end
                         else
                            tx_state <= TX_SEND_DATA;
      endcase
end

//Az USRT interfsz TXD vonalnak meghajtsa.
assign usrt_txd = tx_shr[0];


//******************************************************************************
//* Az adsi FIFO.                                                             *
//******************************************************************************
wire tx_fifo_rd = (tx_state == TX_READ_FIFO);

fifo tx_fifo(
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(tx_fifo_clr),                  //Reset jel
   
   //Adatvonalak.
   .data_in(cpu2dmem_data),            //A FIFO-ba rand adat
   .data_out(tx_fifo_dout),            //A FIFO-bl kiolvasott adat
   
   //Vezrl bemenetek.
   .write(data_reg_wr),                //rs engedlyez jel
   .read(tx_fifo_rd),                  //Olvass engedlyez jel
   
   //Sttusz kimenetek.
   .empty(tx_fifo_empty),              //A FIFO res
   .full(tx_fifo_full)                 //A FIFO tele van
);


endmodule
