//******************************************************************************
//* MiniRISC rendszer top-level modul.                                         *
//******************************************************************************
module minirisc_system(
   //rajel s reset.
   input  wire        clk16M,    //16 MHz rajel
   input  wire        rstbt,     //Reset nyomgomb
   
   //Perifrik.
   input  wire [7:0]  sw,        //DIP kapcsol
   input  wire [3:0]  bt,        //Nyomgombok
   output wire [7:0]  ld,        //LED-ek
   output wire [7:0]  seg_n,     //Szegmens vezrl jelek (aktv alacsony)
   output wire [3:0]  dig_n,     //Digit kivlaszt jelek (aktv alacsony)
   output wire [4:0]  col_n,     //Oszlop kivlaszt jelek (aktv alacsony)
   
   //USRT.
   input  wire        dev_clk,   //USRT rajel
   input  wire        dev_mosi,  //Soros adatbemenet
   output wire        dev_miso,  //Soros adatkimenet
   
   //GPIO.
   inout  wire [12:5] aio,       //A bvtcsatlakoz
   inout  wire [12:5] bio        //B bvtcsatlakoz
);

//******************************************************************************
//* rajel s reset.                                                           *
//******************************************************************************
wire clk = clk16M;
wire rst;


//******************************************************************************
//* MiniRISC CPU v2.0.                                                         *
//******************************************************************************
wire [7:0]  cpu2prgmem_addr;
wire [15:0] prgmem2cpu_data;
wire [7:0]  cpu2datamem_addr;
wire        cpu2datamem_wr;
wire        cpu2datamem_rd;
wire [7:0]  cpu2datamem_data;
wire [7:0]  datamem2cpu_data;
wire        irq;
wire [22:0] dbg2cpu_data;
wire [44:0] cpu2dbg_data;

minirisc_cpu minirisc_cpu(
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rst),                          //Aszinkron reset
   
   //A programmemrival kapcsolatos jelek.
   .cpu2pmem_addr(cpu2prgmem_addr),    //Cmbusz
   .pmem2cpu_data(prgmem2cpu_data),    //Olvassi adatbusz
   
   //Az adatmemrival kapcsolatos jelek.
   .cpu2dmem_addr(cpu2datamem_addr),   //Cmbusz
   .cpu2dmem_wr(cpu2datamem_wr),       //rs engedlyez jel
   .cpu2dmem_rd(cpu2datamem_rd),       //Olvass engedlyez jel
   .cpu2dmem_data(cpu2datamem_data),   //rsi adatbusz
   .dmem2cpu_data(datamem2cpu_data),   //Olvassi adatbusz
   
   //Megszaktskr bemenet (aktv magas szintrzkeny).
   .irq(irq),
   
   //Debug interfsz.
   .dbg2cpu_data(dbg2cpu_data),        //Jelek a debug modultl a CPU fel
   .cpu2dbg_data(cpu2dbg_data)         //Jelek a CPU-tl a debug modul fel
);


//******************************************************************************
//* Debug modul.                                                               *
//******************************************************************************
wire [7:0]  dbg2prgmem_addr;
wire [15:0] dbg2prgmem_data;
wire        dbg2prgmem_wr;

debug_module debug_module(
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst_in(rstbt),                     //Reset bemenet
   .rst_out(rst),                      //Reset jel a rendszer szmra
   
   //A programmemria rshoz szksges jelek.
   .dbg2pmem_addr(dbg2prgmem_addr),    //rsi cm
   .dbg2pmem_data(dbg2prgmem_data),    //A memriba rand adat
   .dbg2pmem_wr(dbg2prgmem_wr),        //rs engedlyez jel
   
   //Debug interfsz a CPU fel.
   .dbg2cpu_data(dbg2cpu_data),        //Jelek a debug modultl a CPU fel
   .cpu2dbg_data(cpu2dbg_data)         //Jelek a CPU-tl a debug modul fel
);


//******************************************************************************
//* 256 x 16 bites programmemria (elosztott RAM).                             *
//******************************************************************************
(* ram_style = "distributed" *)
reg [15:0] prg_mem [255:0];

always @(posedge clk)
begin
   if (dbg2prgmem_wr)
      prg_mem[dbg2prgmem_addr] <= dbg2prgmem_data;
end

assign prgmem2cpu_data = prg_mem[cpu2prgmem_addr];


//******************************************************************************
//* 128 x 8 bites adatmemria.                                                 *
//* Cmtartomny: 0x00 - 0x7F (rhat/olvashat)                               *
//******************************************************************************
(* ram_style = "distributed" *)
reg  [7:0] data_mem [127:0];

wire [6:0] data_mem_addr = cpu2datamem_addr[6:0];
wire       data_mem_wr   = cpu2datamem_wr & ~cpu2datamem_addr[7];
wire       data_mem_rd   = cpu2datamem_rd & ~cpu2datamem_addr[7];
wire [7:0] data_mem_din  = cpu2datamem_data;
wire [7:0] data_mem_dout = (data_mem_rd) ? data_mem[data_mem_addr] : 8'd0;

always @(posedge clk)
begin
   if (data_mem_wr)
      data_mem[data_mem_addr] <= data_mem_din;
end


//******************************************************************************
//* LED perifria.                                                             *
//* Cmtartomny: 0x80 (rhat/olvashat)                                      *
//******************************************************************************
wire [7:0] led_dout;

basic_owr #(
   //A perifria bziscme.
   .BASEADDR(8'h80)
) leds (
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rst),                          //Reset jel
   
   //A busz interfsz jelei.
   .cpu2dmem_addr(cpu2datamem_addr),   //Cmbusz
   .cpu2dmem_wr(cpu2datamem_wr),       //rs engedlyez jel
   .cpu2dmem_rd(cpu2datamem_rd),       //Olvass engedlyez jel
   .cpu2dmem_data(cpu2datamem_data),   //rsi adatbusz
   .dmem2cpu_data(led_dout),           //Olvassi adatbusz
   
   //A GPIO interfsz jelei.
   .gpio_out(ld)                       //Az IO lbakra kirand adat
);


//******************************************************************************
//* DIP kapcsol perifria.                                                    *
//* Cmtartomny: 0x81 (csak olvashat)                                        *
//******************************************************************************
wire [7:0] dipsw_dout;

basic_in #(
   //A perifria bziscme.
   .BASEADDR(8'h81)
) dip_switch (
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rst),                          //Reset jel
   
   //A busz interfsz jelei.
   .cpu2dmem_addr(cpu2datamem_addr),   //Cmbusz
   .cpu2dmem_rd(cpu2datamem_rd),       //Olvass engedlyez jel
   .dmem2cpu_data(dipsw_dout),         //Olvassi adatbusz
   
   //A GPIO interfsz jelei.
   .gpio_in(sw)                        //Az IO lbak aktulis rtke
);


//******************************************************************************
//* Nyomgomb perifria.                                                       *
//* Cmtartomny: 0x82 (csak olvashat)                                        *
//******************************************************************************
wire [7:0] btn_dout;

basic_in #(
   //A perifria bziscme.
   .BASEADDR(8'h82)
) buttons (
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rst),                          //Reset jel
   
   //A busz interfsz jelei.
   .cpu2dmem_addr(cpu2datamem_addr),   //Cmbusz
   .cpu2dmem_rd(cpu2datamem_rd),       //Olvass engedlyez jel
   .dmem2cpu_data(btn_dout),           //Olvassi adatbusz
   
   //A GPIO interfsz jelei.
   .gpio_in({4'd0, bt})                //Az IO lbak aktulis rtke
);


//******************************************************************************
//* GPIO (A bvtcsatlakoz).                                                 *
//* Cmtartomny: 0x84 - 0x87 (rhat/olvashat)                               *
//******************************************************************************
wire [7:0] gpio_a_dout;
wire [7:0] aio_out;
wire [7:0] aio_dir;

basic_io #(
   //A perifria bziscme.
   .BASEADDR(8'h84)
) gpio_a (
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rst),                          //Reset jel
   
   //A busz interfsz jelei.
   .cpu2dmem_addr(cpu2datamem_addr),   //Cmbusz
   .cpu2dmem_wr(cpu2datamem_wr),       //rs engedlyez jel
   .cpu2dmem_rd(cpu2datamem_rd),       //Olvass engedlyez jel
   .cpu2dmem_data(cpu2datamem_data),   //rsi adatbusz
   .dmem2cpu_data(gpio_a_dout),        //Olvassi adatbusz
   
   //A GPIO interfsz jelei.
   .gpio_out(aio_out),                 //Az IO lbakra kirand adat
   .gpio_in(aio),                      //Az IO lbak aktulis rtke
   .gpio_dir(aio_dir)                  //A kimeneti meghajt engedlyez jele
);

//Az A bvtcsatlakoz jeleinek meghajtsa.
genvar i;

generate
   for (i = 0; i < 8; i = i + 1)
   begin: aio_loop
      assign aio[i+5] = (aio_dir[i]) ? aio_out[i] : 1'bz; 
   end
endgenerate


//******************************************************************************
//* GPIO (B bvtcsatlakoz).                                                 *
//* Cmtartomny: 0x88 - 0x8B (rhat/olvashat)                               *
//******************************************************************************
wire [7:0] gpio_b_dout;
wire [7:0] bio_out;
wire [7:0] bio_dir;

basic_io #(
   //A perifria bziscme.
   .BASEADDR(8'h88)
) gpio_b (
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rst),                          //Reset jel
   
   //A busz interfsz jelei.
   .cpu2dmem_addr(cpu2datamem_addr),   //Cmbusz
   .cpu2dmem_wr(cpu2datamem_wr),       //rs engedlyez jel
   .cpu2dmem_rd(cpu2datamem_rd),       //Olvass engedlyez jel
   .cpu2dmem_data(cpu2datamem_data),   //rsi adatbusz
   .dmem2cpu_data(gpio_b_dout),        //Olvassi adatbusz
   
   //A GPIO interfsz jelei.
   .gpio_out(bio_out),                 //Az IO lbakra kirand adat
   .gpio_in(bio),                      //Az IO lbak aktulis rtke
   .gpio_dir(bio_dir)                  //A kimeneti meghajt engedlyez jele
);

//A bvtcsatlakoz jeleinek meghajtsa.
genvar j;

generate
   for (j = 0; j < 8; j = j + 1)
   begin: bio_loop
      assign bio[j+5] = (bio_dir[j]) ? bio_out[j] : 1'bz; 
   end
endgenerate


//******************************************************************************
//* Idzt perifria.                                                         *
//* Cmtartomny: 0x8C - 0x8D (rhat/olvashat)                               *
//******************************************************************************
wire [7:0] timer_dout;
wire       timer_irq;

basic_timer #(
   //A perifria bziscme.
   .BASEADDR(8'h8c)
) timer (
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rst),                          //Reset jel
   
   //A busz interfsz jelei.
   .cpu2dmem_addr(cpu2datamem_addr),   //Cmbusz
   .cpu2dmem_wr(cpu2datamem_wr),       //rs engedlyez jel
   .cpu2dmem_rd(cpu2datamem_rd),       //Olvass engedlyez jel
   .cpu2dmem_data(cpu2datamem_data),   //rsi adatbusz
   .dmem2cpu_data(timer_dout),         //Olvassi adatbusz
   
   //Megszaktskr kimenet.
   .irq(timer_irq)
);


//******************************************************************************
//* Slave USRT perifria.                                                      *
//* Cmtartomny: 0x8E - 0x8F (rhat/olvashat)                               *
//******************************************************************************
wire [7:0] usrt_dout;
wire       usrt_irq;

slave_usrt #(
   //A perifria bziscme.
   .BASEADDR(8'h8e)
) usrt (
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rst),                          //Reset jel
   
   //A soros interfsz jelei.
   .usrt_clk(dev_clk),                 //USRT rajel
   .usrt_rxd(dev_mosi),                //Soros adatbemenet
   .usrt_txd(dev_miso),                //Soros adatkimenet
   
   //A busz interfsz jelei.
   .cpu2dmem_addr(cpu2datamem_addr),   //Cmbusz
   .cpu2dmem_wr(cpu2datamem_wr),       //rs engedlyez jel
   .cpu2dmem_rd(cpu2datamem_rd),       //Olvass engedlyez jel
   .cpu2dmem_data(cpu2datamem_data),   //rsi adatbusz
   .dmem2cpu_data(usrt_dout),          //Olvassi adatbusz
   
   //Megszaktskr kimenet.
   .irq(usrt_irq)
);


//******************************************************************************
//* Kijelz perifria.                                                         *
//* Cmtartomny: 0x90 - 0x9f (rhat/olvashat)                               *
//******************************************************************************
wire [7:0] display_dout;

basic_display #(
   //A perifria bziscme.
   .BASEADDR(8'h90)
) display (
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rst),                          //Reset jel
   
   //A processzorhoz trtn illesztshez szksges jelek.
   .cpu2dmem_addr(cpu2datamem_addr),   //Cmbusz
   .cpu2dmem_wr(cpu2datamem_wr),       //rs engedlyez jel
   .cpu2dmem_rd(cpu2datamem_rd),       //Olvass engedlyez jel
   .cpu2dmem_data(cpu2datamem_data),   //rsi adatbusz
   .dmem2cpu_data(display_dout),       //Olvassi adatbusz
   
   //A kijelzk vezrlshez szksges jelek.
   .seg_n(seg_n),                      //Szegmens vezrl jelek (aktv alacsony)
   .dig_n(dig_n),                      //Digit kivlaszt jelek (aktv alacsony)
   .col_n(col_n)                       //Oszlop kivlaszt jelek (aktv alacsony)
);


//******************************************************************************
//* A processzor olvassi adatbusznak meghajtsa.                             *
//******************************************************************************
assign datamem2cpu_data = data_mem_dout |
                          led_dout      |
                          dipsw_dout    |
                          btn_dout      |
                          gpio_a_dout   |
                          gpio_b_dout   |
                          timer_dout    |
                          usrt_dout     |
                          display_dout;
                          
assign irq = timer_irq | usrt_irq;


endmodule
